@@ -22,6 +22,8 @@ class MockedAPI(API):
                 re.compile(r'tickets\?page=1&per_page=100'): self.read_test_file('all_tickets.json'),
                 re.compile(r'tickets/1$'): self.read_test_file('ticket_1.json'),
                 re.compile(r'tickets/1/conversations'): self.read_test_file('conversations.json'),
+                re.compile(r'companies\?page=1&per_page=100$'): self.read_test_file('companies.json'),
+                re.compile(r'companies/1$'): self.read_test_file('company.json'),
                 re.compile(r'contacts\?page=1&per_page=100$'): self.read_test_file('contacts.json'),
                 re.compile(r'contacts/1$'): self.read_test_file('contact.json'),
                 re.compile(r'customers/1$'): self.read_test_file('customer.json'),
